import 'package:flutter/cupertino.dart';
import 'package:nui_core/nui_core.dart';
import 'package:nui_core/ui/component/nui_shader_mask.dart';
import 'package:nui_core/ui/component/nui_state_widget.dart';
import 'package:pull_to_refresh/pull_to_refresh.dart';

enum NUIListItemAnimation{
  SCALE,
  SLIDE,
  SIZE,
  FADE,
  NONE
}

class NUILoadingListItem {
  NUILoadingListItem({@required this.inflater, this.shadowRadius});
  ListLoadingInflater inflater;
  double shadowRadius;
}

typedef NUIListScrollListener = void Function(double scrollPosition, bool atTop, bool atBottom);
typedef NUIListDataCallback = void Function(dynamic data);
typedef NUIListDataStringClassCallback = void Function(String id, Type classType);
typedef NUIListDataPositionCallback = void Function(int position, dynamic data);
typedef NUIListPositionCallback = void Function(int position);
typedef NUIListDataListCallback = void Function(List<dynamic> data, bool append);
typedef NUIListMapUpdateCallback = void Function(Map<String, dynamic>);
typedef NUIListMapGetCallback = Map<String, dynamic> Function();

class NUIListController{
  NUIListController({this.onLoadComplete});
  Function onLoadComplete;
  NUIListDataCallback _insertItemCallback;
  NUIListDataListCallback _insertItemsCallback;
  NUIListDataPositionCallback _insertItemAtCallback;
  NUIListPositionCallback _removeItemAtCallback;
  NUIListDataCallback _removeItemCallback;
  NUIListDataStringClassCallback _removeItemByIDCallback;
  Function _clearCallback;
  Function _refreshListCallback;
  Function _clearCheckedMapCallback;
  NUIListMapGetCallback _updateCheckedMapCallback;
  NUIListMapGetCallback _updateConfigMapCallback;
  NUIListMapGetCallback _getCheckedMapCallback;
  Function _doneCheckedMapCallback;
  Function _doneConfigMapCallback;
  Function _reloadCallback;

  void _onLoadComplete(){
    if(onLoadComplete != null) onLoadComplete();
  }

  Map<String, dynamic> getCheckedMap(){
    if(_getCheckedMapCallback != null) return _getCheckedMapCallback();
    return Map<String, dynamic>();
  }

  void _dispose(){
    _getCheckedMapCallback = null;
    _insertItemCallback = null;
    _insertItemAtCallback = null;
    _insertItemsCallback = null;
    _removeItemAtCallback = null;
    _clearCheckedMapCallback = null;
    _clearCallback = null;
    _refreshListCallback = null;
    _updateConfigMapCallback = null;
    _updateCheckedMapCallback = null;
    _reloadCallback = null;
    _removeItemCallback = null;
    _removeItemByIDCallback = null;
  }

  void insertItem(dynamic data){
    if(_insertItemCallback != null) _insertItemCallback(data);
  }
  void insertItems(List<dynamic> data, {bool append: true}){
    if(_insertItemsCallback != null) _insertItemsCallback(data, append ?? true);
  }
  void insertItemAt(int position, dynamic data){
    if(_insertItemAtCallback != null) _insertItemAtCallback(position, data);
  }
  void removeItem(dynamic data){
    if(_removeItemCallback != null) _removeItemCallback(data);
  }
  void removeItemAt(int position){
    if(_removeItemAtCallback != null) _removeItemAtCallback(position);
  }
  void removeItemByID(String id, Type classType){
    if(_removeItemByIDCallback != null) _removeItemByIDCallback(id, classType);
  }
  void clear(){
    if(_clearCallback != null) _clearCallback();
  }
  void reload(){
    if(_reloadCallback != null) _reloadCallback();
  }
  void refreshList(){
    if(_refreshListCallback != null) _refreshListCallback();
  }
  void clearCheckedMap(){
    if(_clearCheckedMapCallback != null) _clearCheckedMapCallback();
  }
  void updateCheckedMap(NUIListMapUpdateCallback update){
    if(update != null && _updateCheckedMapCallback != null){
      update(_updateCheckedMapCallback());
      if(_doneCheckedMapCallback != null) _doneCheckedMapCallback();
    }
  }
  void updateConfigMap(NUIListMapUpdateCallback update){
    if(update != null && _updateConfigMapCallback != null){
      update(_updateConfigMapCallback());
      if(_doneConfigMapCallback != null) _doneConfigMapCallback();
    }
  }
}

abstract class NUIAnimatedListInterface<T, C> extends StatefulWidget{
  NUIAnimatedListInterface({
    @required this.repo,
    this.initialList,
    @required this.orientation,
    this.padding,
    this.shrinkWrap,
    this.startDivider,
    this.endDivider,
    this.middleDivider,
    this.fixedFirstItem,
    this.fixedLastItem,
    this.checkedMap,
    this.configMap,
    @required this.clickListener,
    this.skeletalCount,
    this.errorHolder,
    this.placeHolder,
    this.skeletalInflater,
    this.paginationInflater,
    this.scrollPhysics,
    this.enablePagination,
    this.enableSwipeRefresh,
    this.minimumPaginateCount,
    this.paginateMargin,
    this.bloc,
    this.skipAnimation: false,
    this.scrollListener,
    this.itemAnimationDuration,
    this.controller,
    this.shaderType : NUIShaderMaskType.NONE,
    this.itemAnimation : NUIListItemAnimation.SCALE
  });

  final NUIListController controller;
  final NUIListItemRepository repo;
  final List<T> initialList;
  final Axis orientation;
  final EdgeInsets padding;

  final Duration itemAnimationDuration;

  final NUIListScrollListener scrollListener;
  final NUIShaderMaskType shaderType;

  final bool shrinkWrap;
  final NUIListItemAnimation itemAnimation;

  final double startDivider;
  final double middleDivider;
  final double endDivider;

  final ListFixedItemGetter fixedFirstItem;
  final ListFixedItemGetter fixedLastItem;

  final Map<String, dynamic> checkedMap;
  final Map<String, dynamic> configMap;

  //Click callback
  final OnListItemClickListener clickListener;

  //Loader widgets
  final int skeletalCount;

  //Widget functions
  final ListErrorHolderInflater<T, C> errorHolder;
  final ListPlaceHolderInflater placeHolder;
  final NUILoadingListItem skeletalInflater;
  final NUILoadingListItem paginationInflater;

  final ScrollPhysics scrollPhysics;

  final bool enablePagination;
  final int minimumPaginateCount;
  final int paginateMargin;
  final bool enableSwipeRefresh;

  final bool skipAnimation;

  final NUIListViewBloc<T, C> bloc;
}

abstract class NUIAnimatedListState<T, C, V extends NUIAnimatedListInterface<T, C>> extends State<V>{
  static const String _MODULE = "NUIAnimatedListState";
  final GlobalKey<AnimatedListState> listKey = GlobalKey<AnimatedListState>();
  final GlobalKey<SliverAnimatedListState> sliverListKey = GlobalKey<SliverAnimatedListState>();

  final items = List();
  final String instanceId = randomUUID();
  Map<String, dynamic> checkedMap;
  Map<String, dynamic> configMap;
  NUIListItemAnimation itemAnimation;
  RefreshController _refreshController;
  double _offset = 0.0;
  NUIListViewModel<T, C> curModel;
  NUIListViewBloc<T, C> bloc;
  NUIListViewParentHandler actionHandler;

  Widget renderSkeletal(int position) => widget.skeletalInflater != null ? _wrapWithPadding(position, widget.skeletalInflater.inflater(position), additionalPadding: widget.skeletalInflater.shadowRadius ?? 0) : null;
  Widget renderPaginate(int position) => widget.paginationInflater != null ? _wrapWithPadding(position, widget.paginationInflater.inflater(position), additionalPadding: widget.paginationInflater.shadowRadius ?? 0) : null;
  NUIListItemRepository repo() => widget.repo;
  double startDivider() => widget.startDivider ?? 0;
  double middleDivider() => widget.middleDivider ?? 0;
  double endDivider() => widget.endDivider ?? 0;
  Axis orientation() => widget.orientation ?? Axis.vertical;

  Duration animateDuration;

  final PaginationLoad pagination = PaginationLoad(); //Data model for pagination
  final SwipeRefresh swipeRefresh = SwipeRefresh(); //Data model for swipe refresh
  dynamic fixedFirstItem;
  dynamic fixedLastItem;
  NUIListController controller;

  bool _atTop = true;
  bool _atBottom = false;
  NUIShaderMaskType toUseShade = NUIShaderMaskType.NONE;
  RefreshController refreshController() => _refreshController;

  @override
  void initState(){
    super.initState();
    controller = widget.controller ?? NUIListController();
    animateDuration = widget.itemAnimationDuration ?? Duration(milliseconds: 200);

    checkedMap = widget.checkedMap ?? Map();
    configMap = widget.configMap ?? Map();
    itemAnimation = widget.itemAnimation ?? NUIListItemAnimation.SCALE;

    bloc = widget.bloc;
    curModel = NUIListViewModel<T, C>(data: widget.initialList ?? List());
    if(!isNullOrEmpty(widget.initialList)){
      items.addAll(widget.initialList);
    }
    _refreshController = new RefreshController(initialRefresh: false);
    fixedFirstItem = widget.fixedFirstItem != null ? widget.fixedFirstItem() : null;
    fixedLastItem = widget.fixedLastItem != null ? widget.fixedLastItem() : null;

    actionHandler = NUIListViewParentHandler(widget.clickListener, checkedMap, configMap, (){
      bool updated = updateNUIState(id: instanceId, update: items);
    }, (){
      bool updated = updateNUIState(id: instanceId, update: items);
    });

    controller._updateCheckedMapCallback = () => checkedMap;
    controller._updateConfigMapCallback = () => configMap;
    controller._doneCheckedMapCallback = () => updateNUIState(id: instanceId, update: items);
    controller._doneConfigMapCallback = () => updateNUIState(id: instanceId, update: items);
    controller._refreshListCallback = (){
      logNUI("NUIListInterface", "Refreshing list, items size : ${items.length}");
      updateNUIState(id: instanceId, update: items);
    };
    controller._clearCallback = (){
      items.clear();
      updateNUIState(id: instanceId, update: items);
    };
    controller._removeItemAtCallback = (position){
      logNUI("NUIListInterface", "Removing item at position $position");
      removeItemAt(position);
    };
    controller._removeItemCallback = (data){
      removeItem(data);
    };
    controller._removeItemByIDCallback = (id, type){
      removeItemByID(type, id);
    };
    controller._insertItemsCallback = (data, append){
      logNUI("NUIListInterface", "Inserting items to list");
      addItems(data, append: append);
    };
    controller._insertItemAtCallback = (position, data){
      logNUI("NUIListInterface", "Inserting item to position $position");
      addItemAt(data, position);
    };
    controller._insertItemCallback = (data) {
      logNUI("NUIListInterface", "Inserting item to the list");
      addItem(data);
    };
    controller._clearCheckedMapCallback = (){
      checkedMap.clear();
      updateNUIState(id: instanceId, update: items);
    };
    controller._reloadCallback = (){
      _reload();
    };
    controller._getCheckedMapCallback = (){
      return checkedMap;
    };
  }

  bool _hasSkeletalLoad(){
    return widget.skeletalInflater != null && widget.skeletalCount > 0;
  }

  dynamic _getItem(int position){
    return items.getOrNull(position);
  }

  Widget _renderListItem(dynamic data, int position, animation){
    if(data is PaginationLoad){
      final renderedPaginate = renderPaginate(position);
      //Pagination loading
      return renderedPaginate ?? NUIListViewProgressWidget();
    }
    else if(data == null){
      final renderedSkeletal = renderSkeletal(position);
      return renderedSkeletal ?? NUIListViewProgressWidget();
    }

    NUIListItem item = repo().getItemByClass(data.runtimeType);
    if(item == null){
      throwNUI(_MODULE, "List item for type : ${data.runtimeType.toString()} is not registered in the repository");
    }

    NUIListActionController handler = NUIListActionController(actionHandler, NUIListItemModel(position, data, checkedMap: checkedMap, configMap: configMap));
    final shadowRadius = item.shadowRadius ?? 0;
    return _wrapWithPadding(position, item.inflater(data, position, handler), additionalPadding: shadowRadius);
  }

  Widget _wrapWithPadding(int position, Widget child, {double additionalPadding}){
    EdgeInsets padding;
    final addPadding = additionalPadding ?? 0;
    if(orientation() == Axis.vertical){
      padding = EdgeInsets.only(
          top: position == 0 ? startDivider() + addPadding ?? addPadding : addPadding,
          bottom: position < items.length - 1 ? middleDivider() + addPadding ?? addPadding : endDivider() + addPadding ?? addPadding,
          left: addPadding,
          right: addPadding
      );
    }
    else{
      padding = EdgeInsets.only(
          left: position == 0 ? startDivider() + addPadding ?? addPadding : addPadding,
          right: position < items.length - 1 ? middleDivider() + addPadding ?? addPadding : endDivider() + addPadding ?? addPadding,
          top: addPadding,
          bottom: addPadding
      );
    }
    return Container(
      padding: padding,
      child: child,
    );
  }

  Widget renderItem(BuildContext context, int index){
    return _renderListItem(_getItem(index), index, null);
  }

  Widget animateItem(BuildContext context, int index, animation){
    final data = _getItem(index);
    Widget child;
    if(data != null){
      switch(itemAnimation){
        case NUIListItemAnimation.SCALE: child = _scaleItem(context, data, index, animation); break;
        case NUIListItemAnimation.SIZE: child = _sizeItem(context, data, index, animation); break;
        case NUIListItemAnimation.SLIDE: child = _slideItem(context, data, index, animation); break;
        case NUIListItemAnimation.FADE: child = _fadeItem(context, data, index, animation); break;
        default: child = _renderListItem(_getItem(index), index, animation); break;
      }
    }
    else{
      child = _renderListItem(_getItem(index), index, animation);
    }
    return child;
  }

  Widget _fadeItem(BuildContext context, dynamic data, int index, animation, {Widget child}) {
    final widget = child ?? _renderListItem(data, index, animation);
    return FadeTransition(
      opacity: animation,
      child: widget
    );
  }

  Widget _scaleItem(BuildContext context, dynamic data, int index, animation, {Widget child}) {
    final widget = child ?? _renderListItem(data, index, animation);
    return SizeTransition(
      axis: orientation(),
      sizeFactor: animation,
      child: ScaleTransition(
          alignment: Alignment.center,
          scale: animation,
          child: widget
      ),
    );
  }

  Widget _sizeItem(BuildContext context, dynamic data, int index, animation, {Widget child}) {
    final widget = child ?? _renderListItem(data, index, animation);
    return FadeTransition(
      opacity: animation,
      child: SizeTransition(
          axis: orientation(),
          sizeFactor: animation,
          child: widget
      ),
    );
  }

  Widget _slideItem(BuildContext context, dynamic data, int index, animation, {Widget child}) {
    final widget = child ?? _renderListItem(data, index, animation);
    return SlideTransition(
        position: Tween<Offset>(
          begin: const Offset(-1, 0),
          end: Offset(0, 0),
        ).animate(animation),
        child: widget
    );
  }

  void addItem(dynamic data){
    addItemAt(data, items.length);
  }

  void addItemAt(dynamic data, int position){
    listKey.currentState?.insertItem(position, duration: animateDuration);
    sliverListKey.currentState?.insertItem(position, duration: animateDuration);
    items.insert(position, data);
  }

  Future<bool> addItems(List<dynamic> newItems, {bool reassemble: false, bool append: true}) async{
    if(isNullOrEmpty(newItems)) return false;

    if(reassemble == true || widget.skipAnimation == true || itemAnimation == NUIListItemAnimation.NONE){
      if(append != false) {
        items.addAll(newItems);
      }
      else{
        final cloneList = List<dynamic>();
        cloneList.addAll(items);
        items.clear();
        items.addAll(newItems);
        items.addAll(cloneList);
      }
      listKey.currentState?.setState(() {

      });
      sliverListKey.currentState?.setState(() {

      });
    }
    else{
      if(append != false) {
        for (dynamic item in newItems) {
          await Future.delayed(animateDuration);
          addItem(item);
        }
      }
      else{
        int position = fixedFirstItem != null ? 1 : 0;
        for(dynamic item in newItems){
          await Future.delayed(animateDuration);
          addItemAt(item, position);
          position++;
        }
      }
    }
    return true;
  }

  void removeItem(dynamic data){
    int indexToRemove;
    int index = 0;
    for(dynamic item in items){
      if(item == data){
        indexToRemove = index;
        break;
      }
      index++;
    }
    if(indexToRemove != null) {
      removeItemAt(indexToRemove);
    }
  }

  void removeItemByID(Type classType, String id){
    int indexToRemove;
    int index = 0;
    for(dynamic item in items){
      if(item.runtimeType == classType){
        NUIListItem listItem = repo().getItemByClass(classType);
        if(listItem.dataId != null && match(id, listItem.dataId(item))){
          indexToRemove = index;
          break;
        }
      }
      index++;
    }
    if(indexToRemove != null) {
      removeItemAt(indexToRemove);
    }
  }

  void updateItems(){

  }

  void removeItemAt(int position){
    if(position > items.length - 1) return;
    listKey.currentState?.removeItem(position, (_, animation) => animateItem(context, position, animation), duration: animateDuration);
    sliverListKey.currentState?.removeItem(position, (_, animation) => animateItem(context, position, animation), duration: animateDuration);
    items.removeAt(position);
  }

  void _addPaginationIfNotAdded(){
    bool hasPagination = false;
    for(dynamic data in items){
      if(data is PaginationLoad){
        hasPagination = true;
        break;
      }
    }

    if(!hasPagination){
      addItem(pagination);
    }
  }

  bool clearItems(){
    if(items.isNullOrEmpty()) return false;
    final cloneList = List();
    cloneList.addAll(items);
    listKey.currentState?.reassemble();
    sliverListKey.currentState?.reassemble();
    items.clear();
    return true;
  }

  void _onPaginationLoad({bool isFirst, C cursor}){
    if(bloc == null) return;
    if (isFirst == true || (widget.enablePagination && !curModel.isLoading() && curModel?.reachedEnd != true)) {
      if(isFirst != true && curModel.data != null && curModel.data.length < widget.minimumPaginateCount) {
        return;
      }
      curModel.paginationLoading = true;
      bloc.startPaginationLoading(isFirst, curModel.data, cursor, onLoadComplete: controller._onLoadComplete);
    }
  }

  void onSwipeRefresh(){
    if(bloc == null) return;
    if (!curModel.isLoading()) {
      curModel.swipeRefreshing = true;
      bloc.startSwipeRefreshLoading(curModel.data, onLoadComplete: controller._onLoadComplete);
    }
  }

  void _onSkeletalLoad(){
    if(bloc == null) return;
    if (!curModel.isLoading()) {
      curModel.skeletalLoading = true;
      bloc.startSkeletalLoading(onLoadComplete: controller._onLoadComplete);
    }
  }

  void _reload(){
    if (_hasSkeletalLoad()) {
      _onSkeletalLoad();
    }
    else {
      _onPaginationLoad(isFirst: true, cursor: null);
    }
  }

  void onScroll(ScrollMetrics metrics){
    final newAtBottom = metrics.pixels >= metrics.maxScrollExtent && !metrics.outOfRange;
    final newAtTop = metrics.pixels <= metrics.minScrollExtent && metrics.outOfRange;

    if(widget.shaderType != null && widget.shaderType != NUIShaderMaskType.NONE && (newAtBottom != _atBottom || newAtTop != _atTop)) {
      //Scroll position changed
      _atBottom = newAtBottom;
      _atTop = newAtTop;

      bool shouldFadeTop = false;
      bool shouldFadeBottom = false;
      if (widget.shaderType == NUIShaderMaskType.TOP_BOTTOM_FADE) {
        shouldFadeTop = !_atTop;
        shouldFadeBottom = !_atBottom;
      }
      else if (widget.shaderType == NUIShaderMaskType.BOTTOM_FADE) {
        shouldFadeBottom = !_atBottom;
      }
      else if (widget.shaderType == NUIShaderMaskType.TOP_FADE) {
        shouldFadeTop = !_atTop;
      }

      if (shouldFadeTop && shouldFadeBottom)
        toUseShade = NUIShaderMaskType.TOP_BOTTOM_FADE;
      else if (shouldFadeTop)
        toUseShade = NUIShaderMaskType.TOP_FADE;
      else if (shouldFadeBottom)
        toUseShade = NUIShaderMaskType.BOTTOM_FADE;
      else
        toUseShade = NUIShaderMaskType.NONE;

      updateNUIState(id: instanceId, update: items);
    }

    if (metrics.maxScrollExtent == metrics.pixels) {
      _onPaginationLoad(cursor: curModel.cursor);
    }
    additionalScrollListener(metrics);

    if(widget.scrollListener != null){
      widget.scrollListener(metrics.pixels, _atTop, _atBottom);
    }
  }

  void additionalScrollListener(ScrollMetrics metrics){
    //Todo something here i guess
  }

  Widget _buildList({Widget otherView}){
    return NUIStateful<List<dynamic>>(
        id: instanceId,
        initial: items,
        child: (newList){
          return NotificationListener(
            child: _getListViewBuilder(otherView: otherView),
            onNotification: (notification){
              if(notification is ScrollNotification){
                _offset = notification.metrics.pixels;
                onScroll(notification.metrics);
              }
              return true;
            },
          );
        }
    );
  }

  Widget _updateList(){
    final model = curModel;
    //Update checked list and config list if specified
    if(model.updateCheckedMap != null) model.updateCheckedMap(checkedMap);
    logNUI(_MODULE, "Checked keys : ${checkedMap.keys}");
    if(model.updateConfigMap != null) model.updateConfigMap(configMap);

    logNUI(_MODULE, "Setting new list model for list ${T.toString()} with cursor type ${C.toString()}");
    curModel = model;
    _refreshController.refreshCompleted();
    if (model.paginationLoading) {
      // Handle progress state
      _addPaginationIfNotAdded();
      return _buildList();
    }
    else if(model.skeletalLoading){
      clearItems();
      //Add number of skeletal (null) models into the list according to the skeletal count specified
      for(int i = 0; i < widget.skeletalCount; i++){
        addItem(null);
      }
      return _buildList();
    }
    else if(model.success == false && isNullOrEmpty(model.data) && widget.errorHolder != null){
      return customScroll() == true ? _buildList(otherView: getErrorHolder(model)) : getErrorHolder(model);
    }
    else if(isNullOrEmpty(model.data) && widget.placeHolder != null){
      return customScroll() == true ? _buildList(otherView: getPlaceHolder()) : getPlaceHolder();
    }
    else {
      // Handle data state
      if (model.appending != true || items.isNullOrEmpty()) {
        //New set of data
        clearItems();
        final toAddList = List();
        if (fixedFirstItem != null) {
          toAddList.add(fixedFirstItem);
        }
        toAddList.addAll(model.data);
        if (fixedLastItem != null) {
          toAddList.add(fixedLastItem);
        }
        addItems(toAddList);
      }
      else {
        //Appending from current list

        if (items.getOrNull(items.length - 1) == null){
          logNUI("NUIListInterface", "Removing skeletal item before adding new items");
          removeItemAt(items.length - 1);
        }
        if(items.getOrNull(items.length - 1) == fixedLastItem){
          logNUI("NUIListInterface", "Removing fixed last item before adding new items");
          removeItemAt(items.length - 1);
        }
        if(items.getOrNull(items.length - 1) == pagination){
          logNUI("NUIListInterface", "Removing pagination before adding new items");
          removeItemAt(items.length - 1);
        }
        final toAddList = List();
        toAddList.addAll(model.newData);
        if (fixedLastItem != null) {
          toAddList.add(fixedLastItem);
        }
        addItems(toAddList);
      }
      items.remove(pagination);
      return _buildList();
    }
  }

  Widget _buildStreamBuilder(){
    return StreamBuilder(
      initialData: NUIListViewModel<T, C>(paginationLoading: true, data: curModel.data, message: null),
      stream: bloc.stream,
      builder: (BuildContext context, AsyncSnapshot snapshot) {
        NUIListViewModel<T, C> model = snapshot.data;
        curModel = model;
        return _updateList();
      },
    );
  }

  Widget _getListViewBuilder({Widget otherView}){
    final builder = otherView ?? getListView(items.length);
    final parent = getParentList(items.length, builder);
    Widget content;
    if(widget.enableSwipeRefresh == true){
      logNUI("NUIListInterface", "Swipe refresh enabled, wrapping list in refresh configuration, is enabled : ${!curModel.isLoading()}");
      content = RefreshConfiguration(
        headerTriggerDistance: 100,
        child: SmartRefresher(
          enablePullDown: !curModel.isLoading(),
          header: NUISwipeRefreshHeader(
            height: 100,
          ),
          controller: refreshController(),
          onRefresh: onSwipeRefresh,
          child: parent ?? builder,
        ),
      );
    }
    else{
      logNUI("NUIListInterface", "Swipe refresh disabled");
      content = parent ?? builder;
    }

    if(toUseShade != null && toUseShade != NUIShaderMaskType.NONE){
      return NUIShaderMask(child: content, shaderType: toUseShade);
    }

    return content;
  }

  Widget getListView(int initialCount);

  Widget getParentList(int initialCount, Widget innerList){
    return null;
  }

  Widget getErrorHolder(NUIListViewModel<T, C> listModel){
    return widget.errorHolder(listModel);
  }

  Widget getPlaceHolder(){
    return widget.placeHolder();
  }

  bool customScroll(){
    return false;
  }

  @override
  Widget build(BuildContext context) {
    if(bloc != null) {
      if (_hasSkeletalLoad()) {
        _onSkeletalLoad();
      }
      else {
        _onPaginationLoad(isFirst: true, cursor: curModel.cursor);
      }
      return NUIBlocWidget(
          bloc: bloc,
          child: _buildStreamBuilder()
      );
    }
    else{
      return _updateList();
    }
  }

  @override
  void dispose(){
    super.dispose();
    controller?._dispose();
  }
}